import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
# Remove once tf.repeat is available in the tf api
from tensorflow.python.ops.ragged.ragged_util import repeat  # yes, it IS used!

o = tf.newaxis
FLOAT_TYPE = tf.float32


def exporter():
    """Export utility modified from https://stackoverflow.com/a/41895194
    Returns export decorator, __all__ list
    """
    all_ = []
    def decorator(obj):
        all_.append(obj.__name__)
        return obj
    return decorator, all_


export, __all__ = exporter()
__all__ += ['float_type', 'exporter', 'repeat']


@export
def float_type():
    return FLOAT_TYPE


@export
def lookup_axis1(x, indices, fill_value=0):
    """Return values of x at indices along axis 1,
       returning fill_value for out-of-range indices.
    """
    # Save shape of x and flatten
    a, b = x.shape
    x = tf.reshape(x, [-1])

    legal_index = indices < b

    # Convert indices to legal indices in flat array
    indices = tf.dtypes.cast(indices, dtype=tf.int32)
    indices = tf.clip_by_value(indices, 0, b - 1)
    indices = indices + b * tf.range(a)[:, o, o]

    # Do indexing
    result = tf.reshape(tf.gather(x,
                                  tf.reshape(indices, shape=(-1,))),
                        shape=indices.shape)

    # Replace illegal indices with fill_value, cast to float explicitly
    return tf.cast(tf.where(legal_index,
                            result,
                            tf.zeros_like(result) + fill_value),
                   dtype=float_type())


@export
def tf_to_np(x):
    """Convert (list/tuple of) tensors x to numpy"""
    if isinstance(x, (list, tuple)):
        return tuple([tf_to_np(y) for y in x])
    if isinstance(x, np.ndarray):
        return x
    return x.numpy()


@export
def np_to_tf(x):
    """Convert (list/tuple of) arrays x to tensorflow"""
    if isinstance(x, (list, tuple)):
        return tuple([np_to_tf(y) for y in x])
    if isinstance(x, tf.Tensor):
        return x
    return tf.convert_to_tensor(x, dtype=float_type())


@export
def tf_log10(x):
    return tf.math.log(x) / tf.math.log(tf.constant(10, dtype=x.dtype))


@export
def safe_p(ps):
    """Clip probabilities to be in [1e-5, 1 - 1e-5]
    NaNs are replaced by 1e-5.
    """
    ps = tf.where(tf.math.is_nan(ps),
                  tf.zeros_like(ps, dtype=float_type()),
                  tf.cast(ps, dtype=float_type()))
    ps = tf.clip_by_value(ps, 1e-5, 1 - 1e-5)
    return ps


@export
def beta_params(mean, sigma):
    """Convert (p_mean, p_sigma) to (alpha, beta) params of beta distribution
    """
    # From Wikipedia:
    # variance = 1/(4 * (2 * beta + 1)) = 1/(8 * beta + 4)
    # mean = 1/(1+beta/alpha)
    # =>
    # beta = (1/variance - 4) / 8
    # alpha
    b = (1. / (8. * sigma ** 2) - 0.5)
    a = b * mean / (1. - mean)
    return a, b


@export
def beta_binom_pmf(x, n, p_mean, p_sigma):
    """Return probability mass function of beta-binomial distribution.

    That is, give the probability of obtaining x successes in n trials,
    if the success probability p is drawn from a beta distribution
    with mean p_mean and standard deviation p_sigma.

    Implemented using Dirichlet Multinomial distribution which is
    identically the Beta-Binomial distribution when len(beta_pars) == 2
    """
    # TODO: check if the number of successes wasn't reversed in the original
    # code. Should we have [x, n-x] or [n-x, x]?

    beta_pars = tf.stack(beta_params(p_mean, p_sigma), axis=-1)

    # DirichletMultinomial only gives correct output on float64 tensors!
    # Cast inputs to float64 explicitly!
    beta_pars = tf.cast(beta_pars, dtype=tf.float64)
    x = tf.cast(x, dtype=tf.float64)
    n = tf.cast(n, dtype=tf.float64)

    counts = tf.stack([x, n-x], axis=-1)
    res = tfp.distributions.DirichletMultinomial(
        n,
        beta_pars,
        # validate_args=True,
        # allow_nan_stats=False
        ).prob(counts)
    res = tf.cast(res, dtype=float_type())
    return tf.where(tf.math.is_finite(res),
                    res,
                    tf.zeros_like(res, dtype=float_type()))


import flamedisx as fd
import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_probability as tfp
from tqdm import tqdm
import typing as ty

export, __all__ = fd.exporter()


@export
class LogLikelihood:
    param_defaults: ty.Dict[str, float]
    data: pd.DataFrame
    sources: ty.Dict[str, fd.Source]
    mu_iterpolators: ty.Dict[str, ty.Callable]

    def __init__(
            self,
            sources: ty.Dict[str, fd.Source.__class__],
            data: pd.DataFrame,
            free_rates: ty.Union[None, str, ty.Tuple[str]] = None,
            batch_size=10,
            n_trials=int(1e5),
            **common_param_specs):

        param_defaults = dict()

        if free_rates is None:
            free_rates = tuple()
        if isinstance(free_rates, str):
            free_rates = (free_rates,)
        for sn in free_rates:
            if sn not in sources:
                raise ValueError(f"Can't free rate of unknown source {sn}")
            param_defaults[sn + '_rate_multiplier'] = 1

        # Create sources. Have to copy data, it's modified by set_data
        self.sources = {
            sname: sclass(data.copy(), batch_size=batch_size)
            for sname, sclass in sources.items()}
        del sources    # so we don't use it by accident

        for pname in common_param_specs:
            # Check defaults for common parameters are consistent between
            # sources
            defs = [s.defaults[pname] for s in self.sources.values()]
            if len(set([x.numpy() for x in defs])) > 1:
                raise ValueError(
                    f"Inconsistent defaults {defs} for common parameters")
            param_defaults[pname] = defs[0]

        first_source = self.sources[list(self.sources.keys())[0]]

        self.n_batches = first_source.n_batches
        self.data = first_source.data
        self.param_defaults = param_defaults
        self.param_names = list(param_defaults.keys())
        self.mu_itps = {
            sname: s.mu_interpolator(n_trials=n_trials,
                                     data=s.data,
                                     **common_param_specs)
            for sname, s in self.sources.items()}
        # Not used, but useful for mu smoothness diagnosis
        self.param_specs = common_param_specs

    @tf.function
    def log_likelihood(self, ptensor):
        return sum([self._log_likelihood(ptensor, i_batch=i_batch)
                    for i_batch in range(self.n_batches)])

    def minus_ll(self, ptensor):
        return -2 * self.log_likelihood(ptensor)

    def mu(self, ptensor):
        return self._mu(ptensor)

    def _check_ptensor(self, ptensor):
        if not len(ptensor) == len(self.param_names):
            raise ValueError(
                f"Likelihood takes {len(self.param_names)} params "
                f"but you gave {len(ptensor)}")

    def _get_rate_mult(self, sname, ptensor):
        rmname = sname + '_rate_multiplier'
        if rmname in self.param_names:
            return ptensor[self._param_i(rmname)]
        return 1.

    def _source_kwargs(self, ptensor):
        """Return {param: value} dictionary with keyword arguments
        for source, with values extracted from ptensor"""
        return {pname: ptensor[self._param_i(pname)]
                for pname in self.param_names
                if not pname.endswith('_rate_multiplier')}

    def _param_i(self, pname):
        """Return index of parameter pname"""
        return self.param_names.index(pname)

    def _mu(self, ptensor):
        self._check_ptensor(ptensor)

        mu = tf.constant(0., dtype=fd.float_type())
        for sname, s in self.sources.items():
            mu += (self._get_rate_mult(sname, ptensor)
                   * self.mu_itps[sname](**self._source_kwargs(ptensor)))
        return mu

    def _log_likelihood(self, ptensor, i_batch=None):
        self._check_ptensor(ptensor)

        first_source = self.sources[list(self.sources.keys())[0]]
        lls = tf.zeros(first_source.n_events(i_batch=i_batch),
                       dtype=fd.float_type())

        for sname, s in self.sources.items():
            lls += (
                self._get_rate_mult(sname, ptensor)
                * s._differential_rate(i_batch, **self._source_kwargs(ptensor)))

        ll = tf.reduce_sum(tf.math.log(lls))

        if i_batch is None or i_batch == 0:
            return -self._mu(ptensor) + ll
        return ll

    def _minus_ll(self, ptensor, i_batch=None):
        return -2 * self._log_likelihood(ptensor, i_batch)

    def guess(self):
        """Return tensor of parameter guesses"""
        return tf.convert_to_tensor(
            list(self.param_defaults.values()),
            fd.float_type())

    def params_to_dict(self, values):
        """Return parameter {name: value} dictionary"""
        values = fd.tf_to_np(values)
        return {k: v
                for k, v in zip(self.param_names, values)}

    def bestfit(self, guess=None,
                optimizer=tfp.optimizer.lbfgs_minimize,
                llr_tolerance=0.01,
                get_lowlevel_result=False, **kwargs):
        """Return best-fit parameter tensor

        :param guess: Guess parameters: array or tensor of same length
        as param_names.
        If omitted, will use guess from source defaults.
        :param llr_tolerance: stop minimizer if change in -2 log likelihood
        becomes less than this (roughly: using guess to convert to
        relative tolerance threshold)
        """
        if guess is None:
            guess = self.guess()
        guess = fd.np_to_tf(guess)

        # Unfortunately we can only set the relative tolerance for the
        # objective; we'd like to set the absolute one.
        # Use the guess log likelihood to normalize;
        if llr_tolerance is not None:
            kwargs.setdefault('f_relative_tolerance',
                              llr_tolerance/self.minus_ll(guess))

        # Minimize multipliers to the guess, rather than the guess itself
        # This is a basic kind of standardization that helps make the gradient
        # vector reasonable.
        x_norm = tf.ones(len(guess), dtype=fd.float_type())

        @tf.function
        def objective(x_norm):
            y = tf.constant(0, dtype=fd.float_type())
            grad = tf.constant(0, dtype=fd.float_type())
            for i_batch in range(self.n_batches):
                with tf.GradientTape() as t:
                    t.watch(x_norm)
                    y += self._minus_ll(x_norm * guess, i_batch=i_batch)
                    grad += t.gradient(y, x_norm)
            return y, grad

        res = optimizer(objective, x_norm, **kwargs)
        if get_lowlevel_result:
            return res
        if res.failed:
            raise ValueError(f"Optimizer failure! Result: {res}")
        return res.position * guess

    def inverse_hessian(self, params, save_ram=True):
        """Return inverse hessian (square numpy matrix)
        of -2 log_likelihood at params
        """
        # I could only get higher-order derivatives to work
        # after splitting the parameter vector in separate variables,
        # and using the un-@tf.function'ed likelihood.
        #
        # Tensorflow has tf.hessians, but:
        # https://github.com/tensorflow/tensorflow/issues/29781

        n = len(self.param_names)
        hessian = np.zeros((n, n))

        if save_ram:
            # Evaluate likelihood separately for each derivative.
            for i_batch in tqdm(range(self.n_batches),
                                desc='Computing hessian'):
                for i1 in range(n):
                    for i2 in range(n):
                        if i2 > i1:
                            continue

                        xc = [tf.constant(q) for q in fd.tf_to_np(params)]
                        with tf.GradientTape(persistent=True) as t2:
                            t2.watch(xc[i2])
                            with tf.GradientTape() as t:
                                t.watch(xc[i1])
                                ptensor = tf.stack(xc)
                                y = self._minus_ll(ptensor, i_batch=i_batch)
                            grad = t.gradient(y, xc[i1])
                            hessian[i1, i2] += t2.gradient(grad, xc[i2]).numpy()
                        del t2

            for i1 in range(n):
                for i2 in range(n):
                    if i2 > i1:
                        hessian[i1, i2] = hessian[i2, i1]

        else:
            # Faster, RAM-guzzling algorithm
            # Do a single computation, tracing all the variables.
            # TODO: take advantage of symmetry! Currently 2x wastage!
            for i_batch in tqdm(range(self.n_batches),
                                desc='Computing hessian'):
                xc = [tf.Variable(q) for q in fd.tf_to_np(params)]
                with tf.GradientTape(persistent=True) as t2:
                    with tf.GradientTape(persistent=True) as t:
                        ptensor = tf.stack(xc)
                        y = self._minus_ll(ptensor, i_batch=i_batch)
                    grads = [t.gradient(y, q) for q in xc]
                hessian += np.vstack(
                    [np.array([t2.gradient(g, x)
                               for x in xc])
                    for g in grads])

        return np.linalg.inv(hessian)

    def summary(self, bestfit, inverse_hessian=None, precision=3):
        """Print summary information about best fit"""
        if inverse_hessian is None:
            inverse_hessian = self.inverse_hessian(bestfit)
        stderr, cov = cov_to_std(inverse_hessian)
        for i, pname in enumerate(self.param_names):
            template = "{pname}: {x:.{precision}g} +- {xerr:.{precision}g}"
            print(template.format(
                pname=pname,
                x=bestfit[i],
                xerr=stderr[i],
                precision=precision))

        df = pd.DataFrame(
            {p1: {p2: cov[i1, i2]
                  for i2, p2 in enumerate(self.param_names)}
             for i1, p1 in enumerate(self.param_names)},
            columns=self.param_names)

        # Get rows in the correct order
        df['index'] = [self.param_names.index(x)
                       for x in df.index.values]
        df = df.sort_values(by='index')
        del df['index']

        print("Correlation matrix:")
        pd.set_option('precision', 3)
        print(df)
        pd.reset_option('precision')


@export
def cov_to_std(cov):
    """Return (std errors, correlation coefficent matrix)
    given covariance matrix cov
    """
    std_errs = np.diag(cov) ** 0.5
    corr = cov * np.outer(1 / std_errs, 1 / std_errs)
    return std_errs, corr



"""XENON1T SR0 implementation

"""
import numpy as np
import tensorflow as tf

from multihist import Hist1d
import wimprates

import flamedisx as fd

export, __all__ = fd.exporter()


##
# Electron probability
##

def p_el_thesis(e_kev, a=15, b=-27.7, c=32.5, e0=5.):
    eps = fd.tf_log10(e_kev / e0 + 1e-9)
    qy = a * eps ** 2 + b * eps + c
    return qy * 13.8e-3


def p_el_sr0(e_kev):
    """Return probability of created quanta to become an electron
    for different deposited energies e_kev.

    This uses the charge yield model for XENON1T SR0,
    as published in https://arxiv.org/abs/1902.11297
    (median posterior).
    """
    e_kev = tf.convert_to_tensor(e_kev, dtype=fd.float_type())

    # Parameters from Table II, for SR0
    mean_nexni = 0.15
    w_bbf = 13.8e-3
    q0 = 1.13
    q1 = 0.47
    gamma_er = 0.124 / 4
    omega_er = 31
    F = 120
    delta_er = 0.24

    with np.warnings.catch_warnings():
        np.warnings.filterwarnings('ignore')

        fi = 1 / (1 + mean_nexni)
        nq = e_kev / w_bbf
        ni, nex = nq * fi, nq * (1 - fi)
        wiggle_er = gamma_er * tf.exp(-e_kev / omega_er) * F ** (-delta_er)
        r_er = 1 - tf.math.log(1 + ni * wiggle_er) / (ni * wiggle_er)
        r_er /= (1 + tf.exp(-(e_kev - q0) / q1))
        p_el = ni * (1 - r_er) / nq

    # placeholder value for e = 0 (better than NaN)
    p_el = tf.where(e_kev == 0, tf.ones_like(p_el), p_el)

    return p_el


def p_electron_nr(
        nq,
        alpha=1.280, zeta=0.045, beta=273 * .9e-4,
        gamma=0.0141, delta=0.062,
        drift_field=120):
    # From lenardo et al global fit
    # TODO: account for Penning quenching in photon detection efficiency
    # TODO: so to make field pos-dependent, override this entire f?
    # could be made easier...

    # prevent /0  # TODO can do better than this
    nq = nq + 1e-9

    # Note: final term depends on nq now, not energy
    # this means beta is different from lenardo et al
    nexni = alpha * drift_field ** -zeta * (1 - tf.exp(-beta * nq))
    ni = nq * 1 / (1 + nexni)

    # Fraction of ions NOT participating in recombination
    squiggle = gamma * drift_field ** -delta
    fnotr = tf.math.log(1 + ni * squiggle) / (ni * squiggle)

    # Finally, number of electrons produced..
    n_el = ni * fnotr

    return n_el / nq


##
# Yield maps
##


s1_map, s2_map = [
    fd.InterpolatingMap(fd.get_resource(fd.pax_file(x)))
    for x in ('XENON1T_s1_xyz_ly_kr83m_SR0_pax-642_fdc-AdCorrTPF.json',
              'XENON1T_s2_xy_ly_SR0_24Feb2017.json')]


##
# Flamedisx sources
##


@export
class SR0Source:
    extra_needed_columns = ('x_observed', 'y_observed',
                            'x', 'y', 'z')

    @staticmethod
    def add_extra_columns(d):
        d['s2_relative_ly'] = s2_map(
            np.transpose([d['x_observed'].values,
                          d['y_observed'].values]))
        d['s1_relative_ly'] = s1_map(
            np.transpose([d['x'].values,
                          d['y'].values,
                          d['z'].values]))

    @staticmethod
    def electron_detection_eff(drift_time,
                               *, elife=452e3, extraction_eff=0.96):
        return extraction_eff * tf.exp(-drift_time / elife)

    @staticmethod
    def electron_gain_mean(s2_relative_ly,
                           g2=11.4 / (1 - 0.63) / 0.96):
        return g2 * s2_relative_ly

    electron_gain_std = 11.4 * 0.25 / (1 - 0.63)

    @staticmethod
    def photon_detection_eff(s1_relative_ly,
                             mean_eff=0.142 / (1 + 0.219)):
        return mean_eff * s1_relative_ly


@export
class SR0ERSource(SR0Source, fd.ERSource):

    @staticmethod
    def p_electron(nq, erqy_a=15, erqy_b=-27.7, erqy_c=32.5, erqy_e0=5.):
        return fd.safe_p(p_el_thesis(nq * 13.8e-3,
                                     a=erqy_a, b=erqy_b, c=erqy_c,
                                     e0=erqy_e0))

    @staticmethod
    def p_electron_fluctuation(nq):
        # q3 = 1.7 keV ~= 123 quanta
        return tf.clip_by_value(0.041 * (1. - tf.exp(-nq / 123.)),
                                1e-4,
                                float('inf'))


# Compute events/bin spectrum for a WIMP
example_wimp_es = np.geomspace(1, 50, 100)
example_wimp_rs = wimprates.rate_wimp_std(
    example_wimp_es,
    mw=1e3, sigma_nucleon=1e-45)
example_sp = Hist1d.from_histogram(
    example_wimp_rs[:-1] * np.diff(example_wimp_es),
    example_wimp_es)


@export
class SR0NRSource(SR0Source, fd.NRSource):

    def energy_spectrum(self, drift_time):
        n_evts = len(drift_time)
        return (
            example_sp.bin_centers[np.newaxis,:].repeat(n_evts, axis=0),
            example_sp.histogram[np.newaxis,:].repeat(n_evts, axis=0))

    @staticmethod
    def p_electron(nq):
        return fd.safe_p(p_electron_nr(nq))



import inspect
import typing

import numpy as np
import pandas as pd
from scipy import stats
import tensorflow as tf
import tensorflow_probability as tfp

from tqdm import tqdm

import flamedisx as fd
export, __all__ = fd.exporter()

quanta_types = 'photon', 'electron'
signal_name = dict(photon='s1', electron='s2')

# Data methods that take an additional positional argument
special_data_methods = [
    'p_electron',
    'p_electron_fluctuation',
    'electron_acceptance',
    'photon_acceptance',
    's1_acceptance',
    's2_acceptance',
    'penning_quenching_eff'
]

data_methods = (
    special_data_methods
    + ['energy_spectrum', 'work', 'double_pe_fraction'])
hidden_vars_per_quanta = 'detection_eff gain_mean gain_std'.split()
for _qn in quanta_types:
    data_methods += [_qn + '_' + x for x in hidden_vars_per_quanta]

o = tf.newaxis

@export
class Source:
    data_methods = tuple(data_methods)
    special_data_methods = tuple(special_data_methods)

    # Whether or not to simulate overdispersion in electron/photon split
    # (e.g. due to non-binomial recombination fluctuation)
    do_pel_fluct = True

    # tuple with columns needed from data to run add_extra_columns
    extra_needed_columns = tuple()

    _params: dict = None
    data: pd.DataFrame

    _tensor_cache: typing.Dict[str, tf.Tensor]
    _tensor_cache_list: typing.List[typing.Dict[str, tf.Tensor]]

    def __init__(self,
                 data,
                 batch_size=10,
                 max_sigma=3,
                 data_is_annotated=False,
                 _skip_tf_init=False,
                 _skip_bounds_computation=False,
                 **params):
        self.max_sigma = max_sigma
        self._params = params
        self.data = data

        # Discover which functions need which arguments / dimensions
        # Discover possible parameters
        self.f_dims = {x: [] for x in self.data_methods}
        self.f_params = {x: [] for x in self.data_methods}
        self.defaults = dict()
        for fname in self.data_methods:
            f = getattr(self, fname)
            if not callable(f):
                # Constant
                continue
            for i, (pname, p) in enumerate(
                    inspect.signature(f).parameters.items()):
                if p.default == inspect.Parameter.empty:
                    if not (fname in self.special_data_methods and i == 0):
                        # It's an observable dimension
                        self.f_dims[fname].append(pname)
                else:
                    # It's a parameter that can be fitted
                    self.f_params[fname].append(pname)
                    if (pname in self.defaults
                            and p.default != self.defaults[pname]):
                        raise ValueError(f"Inconsistent defaults for {pname}")
                    self.defaults[pname] = tf.convert_to_tensor(
                        p.default, dtype=fd.float_type())

        if batch_size is None:
            batch_size = len(data)
        self.batch_size = batch_size
        self.n_batches = np.ceil(
            self.n_events() / self.batch_size).astype(np.int)

        if not data_is_annotated:
            self._annotate(_skip_bounds_computation=_skip_bounds_computation)
        if not _skip_tf_init:
            self._populate_tensor_cache()

    def _populate_tensor_cache(self):
        self._tensor_cache = {
            x: fd.np_to_tf(self.data[x].values)
            for x in self.data.columns
            if (np.issubdtype(self.data[x].dtype, np.integer)
                or np.issubdtype(self.data[x].dtype, np.floating))}

        self.dimsizes = dict()
        for var in ['nq',
                  'photon_detected',
                  'electron_detected',
                  'photon_produced',
                  'electron_produced']:
            ma = self._fetch(var + '_max')
            mi = self._fetch(var + '_min')
            self.dimsizes[var] = int(tf.reduce_max(ma - mi + 1).numpy())

        # Split up tensors for batched evaluation
        start = np.arange(self.n_batches) * self.batch_size
        stop = np.concatenate([start[1:], [self.n_events()]])
        self.tensor_cache_list = [
            {k: v[start[i]:stop[i]]
             for k, v in self._tensor_cache.items()}
            for i in range(self.n_batches)]

    def n_events(self, i_batch=None):
        if i_batch is None:
            return len(self.data)
        if i_batch == self.n_batches - 1:
            return self.n_events() - self.batch_size * (self.n_batches - 1)
        return self.batch_size

    def _fetch(self, x, i_batch=None):
        """Return a tensor column from the original dataframe (self.data)
        :param x: column name
        :param i_batch: Batch index. If None, return results for the entire
        dataset.
        """
        if not hasattr(self, '_tensor_cache'):
            # We're inside annotate, just return the column
            return fd.np_to_tf(self.data[x].values)
        if i_batch is None:
            return self._tensor_cache.get(
                x,
                fd.np_to_tf(self.data[x].values))
        else:
            return self.tensor_cache_list[i_batch][x]

    def gimme(self, fname, bonus_arg=None, i_batch=None, numpy_out=False):
        """Evaluate the model function fname with all required arguments

        :param fname: Name of the model function to compute
        :param bonus_arg: If fname takes a bonus argument, the data for it
        :param numpy_out: If True, return (tuple of) numpy arrays,
        otherwise (tuple of) tensors.
        :param i_batch: Batch index. If None, return results for the entire
        dataset.

        Before using gimme, you must use set_data to
        populate the internal caches.
        """
        # TODO: make a clean way to keep track of i_batch or have it as input
        assert (bonus_arg is not None) == (fname in self.special_data_methods)

        f = getattr(self, fname)

        if callable(f):
            args = [self._fetch(x, i_batch) for x in self.f_dims[fname]]
            if bonus_arg is not None:
                args = [bonus_arg] + args
            kwargs = {pname: self._params.get(pname, self.defaults[pname])
                      for pname in self.f_params[fname]}
            res = f(*args, **kwargs)

        else:
            if bonus_arg is None:
                x = tf.ones(self.n_events(i_batch),
                            dtype=fd.float_type())
            else:
                x = tf.ones_like(bonus_arg, dtype=fd.float_type())
            res = f * x

        if numpy_out:
            return fd.tf_to_np(res)
        return fd.np_to_tf(res)

    def _q_det_clip_range(self, qn):
        return (self.min_s1_photons_detected if qn == 'photon'
                else self.min_s2_electrons_detected,
                None)

    @classmethod
    def annotate_data(cls, data, **params):
        """Add columns to data with inference information"""
        return cls(data, _skip_tf_init=True, **params)

    def _annotate(self, _skip_bounds_computation=False):
        """Annotate self.data with columns needed for inference.
        """
        d = self.data
        self.add_extra_columns(d)

        # Annotate data with eff, mean, sigma
        # according to the nominal model
        for qn in quanta_types:
            for parname in hidden_vars_per_quanta:
                fname = qn + '_' + parname
                try:
                    d[fname] = self.gimme(fname, numpy_out=True)
                except Exception:
                    print(fname)
                    raise
        d['double_pe_fraction'] = self.gimme('double_pe_fraction',
                                             numpy_out=True)

        if _skip_bounds_computation:
            return

        # Find likely number of detected quanta
        obs = dict(photon=d['s1'], electron=d['s2'])
        for qn in quanta_types:
            n_det_mle = (obs[qn] / d[qn + '_gain_mean'])
            if qn == 'photon':
                n_det_mle /= (1 + d['double_pe_fraction'])
            d[qn + '_detected_mle'] = n_det_mle.round().astype(np.int).clip(
                *self._q_det_clip_range(qn))

        # The Penning quenching depends on the number of produced
        # photons.... But we don't have that yet.
        # Thus, "rewrite" penning eff vs detected photons
        # using interpolation
        # TODO: this will fail when someone gives penning quenching some
        # data-dependent args
        _nprod_temp = np.logspace(-1., 8., 1000)
        peff = self.gimme('penning_quenching_eff', _nprod_temp, numpy_out=True)
        d['penning_quenching_eff_mle'] = np.interp(
            d['photon_detected_mle'] / d['photon_detection_eff'],
            _nprod_temp * peff,
            peff)

        # Approximate energy reconstruction (visible energy only)
        # TODO: how to do CES estimate if someone wants a variable W?
        d['nq_vis_mle'] = (
            d['electron_detected_mle'] / d['electron_detection_eff']
            + (d['photon_detected_mle'] / d['photon_detection_eff']
               / d['penning_quenching_eff_mle']))
        d['e_vis'] = self.gimme('work', numpy_out=True) * d['nq_vis_mle']
        d['fel_mle'] = self.gimme('p_electron', d['nq_vis_mle'].values,
                                  numpy_out=True)

        # Find plausble ranges for detected and observed quanta
        # based on the observed S1 and S2 sizes
        # (we could also derive ranges assuming the CES reconstruction,
        #  but these won't work well for outliers along one of the dimensions)
        # TODO: Meh, think about this, considering also computation cost
        # / space width
        for qn in quanta_types:
            # We need the copy, otherwise the in-place modification below
            # will have the side effect of messing up the dataframe column!
            eff = d[qn + '_detection_eff'].values.copy()
            if qn == 'photon':
                eff *= d['penning_quenching_eff_mle'].values

            n_prod_mle = d[qn + '_produced_mle'] = (
                    d[qn + '_detected_mle'] / eff).astype(np.int)

            # Prepare for bounds computation
            n = d[qn + '_detected_mle'].values
            m = d[qn + '_gain_mean'].values
            s = d[qn + '_gain_std'].values
            if qn == 'photon':
                _, scale = self.dpe_mean_std(n, d['double_pe_fraction'],
                                             m, s)
                scale = scale.values
            else:
                scale = n ** 0.5 * s / m

            for bound, sign in (('min', -1), ('max', +1)):
                # For detected quanta the MLE is quite accurate
                # (since fluctuations are tiny)
                # so let's just use the relative error on the MLE
                d[qn + '_detected_' + bound] = stats.norm.ppf(
                    stats.norm.cdf(sign * self.max_sigma),
                    loc=n,
                    scale=scale,
                ).round().clip(*self._q_det_clip_range(qn)).astype(np.int)

                # For produced quanta, it is trickier, since the number
                # of detected quanta is also uncertain.
                # TODO: where did this derivation come from again?
                # TODO: maybe do a second bound based on CES
                q = 1 / eff
                d[qn + '_produced_' + bound] = stats.norm.ppf(
                    stats.norm.cdf(sign * self.max_sigma),
                    loc=n_prod_mle,
                    scale=(q + (q**2 + 4 * n_prod_mle * q)**0.5)/2
                ).round().clip(*self._q_det_clip_range(qn)).astype(np.int)

        # Bounds on total visible quanta
        d['nq_min'] = d['photon_produced_min'] + d['electron_produced_min']
        d['nq_max'] = d['photon_produced_max'] + d['electron_produced_max']

    @staticmethod
    def add_extra_columns(data):
        """Add additional columns to data

        You must add any columns from data you use here to
        extra_needed.columns.

        :param data: pandas DataFrame
        """
        pass

    def batched_differential_rate(self, progress=True, **params):
        progress = (lambda x: x) if not progress else tqdm
        return np.concatenate([
            fd.tf_to_np(self.differential_rate(i_batch=i_batch, **params))
            for i_batch in progress(range(self.n_batches))])

    @tf.function
    def differential_rate(self, i_batch=None, **params):
        return self._differential_rate(i_batch=None, **params)

    def _differential_rate(self, i_batch=None, **params):
        self._params = params
        # (n_events, |photons_produced|, |electrons_produced|)
        y = self.rate_nphnel(i_batch)

        p_ph = self.detection_p('photon', i_batch)
        p_el = self.detection_p('electron', i_batch)
        d_ph = self.detector_response('photon', i_batch)
        d_el = self.detector_response('electron', i_batch)

        # Rearrange dimensions so we can do a single matrix mult
        p_el = tf.transpose(p_el, (0, 2, 1))
        d_ph = d_ph[:, o, :]
        d_el = d_el[:, :, o]
        y = d_ph @ p_ph @ y @ p_el @ d_el
        return tf.reshape(y, [-1])

    def rate_nq(self, nq_1d, i_batch=None):
        """Return differential rate at given number of produced quanta
        differs for ER and NR"""
        # TODO: this implementation echoes that for NR, but I feel there
        # must be a less clunky way...

        # (n_events, |ne|) tensors
        es, rate_e = self.gimme('energy_spectrum', i_batch=i_batch)
        q_produced = tf.cast(
            tf.floor(es / self.gimme('work', i_batch=i_batch)[:, o]),
            dtype=fd.float_type())

        # (n_events, |nq|, |ne|) tensor giving p(nq | e)
        p_nq_e = tf.cast(tf.equal(nq_1d[:, :, o], q_produced[:, o, :]),
                         dtype=fd.float_type())

        q = tf.reduce_sum(p_nq_e * rate_e[:, o, :], axis=2)
        return q

    def rate_nphnel(self, i_batch=None):
        """Return differential rate tensor
        (n_events, |photons_produced|, |electrons_produced|)
        """
        # Get differential rate and electron probability vs n_quanta
        # these four are (n_events, |nq|) tensors
        _nq_1d = self.domain('nq', i_batch)
        rate_nq = self.rate_nq(_nq_1d, i_batch=i_batch)
        pel = self.gimme('p_electron', _nq_1d, i_batch=i_batch)
        pel_fluct = self.gimme('p_electron_fluctuation', _nq_1d, i_batch=i_batch)

        # Create tensors with the dimensions of our fin al result
        # i.e. (n_events, |photons_produced|, |electrons_produced|),
        # containing:
        # ... numbers of photons and electrons produced:
        nph, nel = self.cross_domains('photon_produced', 'electron_produced', i_batch)
        # ... numbers of total quanta produced
        nq = nel + nph
        # ... indices in nq arrays
        _nq_ind = nq - self._fetch('nq_min', i_batch=i_batch)[:, o, o]
        # ... differential rate
        rate_nq = fd.lookup_axis1(rate_nq, _nq_ind)
        # ... probability of a quantum to become an electron
        pel = fd.lookup_axis1(pel, _nq_ind)
        # ... probability fluctuation
        pel_fluct = fd.lookup_axis1(pel_fluct, _nq_ind)
        # Finally, the main computation is simple:
        pel_num = tf.where(tf.math.is_nan(pel),
                           tf.zeros_like(pel, dtype=fd.float_type()),
                           pel)
        pel_clip = tf.clip_by_value(pel_num, 1e-6, 1. - 1e-6)
        pel_fluct_clip = tf.clip_by_value(pel_fluct, 1e-6, 1.)
        if self.do_pel_fluct:
            return rate_nq * fd.beta_binom_pmf(
                nel,
                n=nq,
                p_mean=pel_clip,
                p_sigma=pel_fluct_clip)
        else:
            return rate_nq * tfp.distributions.Binomial(
                total_count=nq, probs=pel_clip).prob(nel)

    def detection_p(self, quanta_type, i_batch=None):
        """Return (n_events, |detected|, |produced|) te nsor
        encoding P(n_detected | n_produced)
        """
        n_det, n_prod = self.cross_domains(quanta_type + '_detected',
                                           quanta_type + '_produced',
                                           i_batch)

        p = self.gimme(quanta_type + '_detection_eff',
                        i_batch=i_batch)[:, o, o]
        if quanta_type == 'photon':
            # Note *= doesn't work, p will get reshaped
            p = p * self.gimme('penning_quenching_eff', n_prod,
                                i_batch=i_batch)

        result = tfp.distributions.Binomial(
                total_count=n_prod,
                probs=tf.cast(p, dtype=fd.float_type())
            ).prob(n_det)
        return result * self.gimme(quanta_type + '_acceptance', n_det,
                                    i_batch = i_batch)

    def domain(self, x, i_batch=None):
        """Return (n_events, |possible x values|) matrix containing all possible integer
        values of x for each event"""
        result1 = tf.cast(tf.range(self.dimsizes[x]),
                          dtype=fd.float_type())[o, :]
        result2 = self._fetch(x + '_min', i_batch=i_batch)[:, o]
        return result1 + result2

    def cross_domains(self, x, y, i_batch=None):
        """Return (x, y) two-tuple of (n_events, |x|, |y|) tensors
        containing possible integer values of x and y, respectively.
        """
        # TODO: somehow mask unnecessary elements and save computation time
        x_size = self.dimsizes[x]
        y_size = self.dimsizes[y]
        # Change to tf.repeat once it's in the api
        result_x = fd.repeat(self.domain(x, i_batch)[:, :, o], y_size, axis=2)
        result_y = fd.repeat(self.domain(y, i_batch)[:, o, :], x_size, axis=1)
        return result_x, result_y

    def detector_response(self, quanta_type, i_batch=None):
        """Return (n_events, |n_detected|) probability of observing the S[1|2]
        for different number of detected quanta.
        """
        ndet = self.domain(quanta_type + '_detected', i_batch)

        observed = self._fetch(
            signal_name[quanta_type], i_batch=i_batch)[:, o]

        # Lookup signal gain mean and std per detected quanta
        mean_per_q = self.gimme(quanta_type + '_gain_mean', i_batch=i_batch)[:, o]
        std_per_q = self.gimme(quanta_type + '_gain_std',i_batch=i_batch)[:, o]

        if quanta_type == 'photon':
            mean, std = self.dpe_mean_std(
                ndet=ndet,
                p_dpe=self.gimme('double_pe_fraction',i_batch=i_batch)[:, o],
                mean_per_q=mean_per_q,
                std_per_q=std_per_q)
        else:
            mean = ndet * mean_per_q
            std = ndet**0.5 * std_per_q

        # add offset to std to avoid NaNs from norm.pdf if std = 0
        result = tfp.distributions.Normal(
                loc=mean, scale=std + 1e-10
            ).prob(observed)

        # Add detection/selection efficiency
        result *= self.gimme(signal_name[quanta_type] + '_acceptance',
                             observed, i_batch=i_batch)
        return result

    @staticmethod
    def dpe_mean_std(ndet, p_dpe, mean_per_q, std_per_q):
        """Return (mean, std) of S1 signals
        :param ndet: photons detected
        :param p_dpe: double pe emission probability
        :param mean_per_q: gain mean per PE
        :param std_per_q: gain std per PE
        """
        npe_mean = ndet * (1 + p_dpe)
        mean = npe_mean * mean_per_q

        # Variance due to PMT resolution
        var = (npe_mean ** 0.5 * std_per_q)**2
        # Variance due to binomial variation in double-PE emission
        var += ndet * p_dpe * (1 - p_dpe)

        return mean, var**0.5

    ##
    # Simulation
    #
    # These are class methods. We could have implemented them as instance
    # methods, but then we'd need a set_data, keep track of state, etc.
    ##

    @classmethod
    def simulate(cls, energies, data=None, **params):
        """Simulate events at energies,
        drawing values of additional observables (e.g. positions)
        from data.

        Will not return | energies | events lost due to
        selection/detection efficiencies
        """
        # simulate_es cannot be a class method; the energy-spectrum might
        # be position/time/other dependent.
        s = cls(data=data, _skip_tf_init=True, **params)
        if isinstance(energies, (float, int)):
            energies = s.simulate_es(int(energies))

        # Create and set new dataset, with just the dimensions we need
        # (note we should NOT include s1 and s2 here, we're going to simulate
        # them)
        d = data[list(set(sum(
            s.f_dims.values(),
            list(s.extra_needed_columns))))]
        d = d.sample(n=len(energies), replace=True)
        s = cls(data=d,
                _skip_tf_init=True,
                _skip_bounds_computation=True,
                **params)
        assert 'e_vis' not in d.columns

        def gimme(*args):
            return s.gimme(*args, numpy_out=True)

        d['energy'] = energies
        s._simulate_nq()

        d['p_el_mean'] = gimme('p_electron', d['nq'].values)
        d['p_el_fluct'] = gimme('p_electron_fluctuation', d['nq'].values)

        d['p_el_actual'] = stats.beta.rvs(
            *fd.beta_params(d['p_el_mean'], d['p_el_fluct']))
        d['p_el_actual'] = np.nan_to_num(d['p_el_actual']).clip(0, 1)
        d['electron_produced'] = stats.binom.rvs(
            n=d['nq'],
            p=d['p_el_actual'])
        d['photon_produced'] = d['nq'] - d['electron_produced']

        d['electron_detected'] = stats.binom.rvs(
            n=d['electron_produced'],
            p=gimme('electron_detection_eff'))
        d['photon_detected'] = stats.binom.rvs(
            n=d['photon_produced'],
            p=(gimme('photon_detection_eff')
               * gimme('penning_quenching_eff', d['photon_produced'].values)))

        d['s2'] = stats.norm.rvs(
            loc=d['electron_detected'] * gimme('electron_gain_mean'),
            scale=d['electron_detected']**0.5 * gimme('electron_gain_std'))

        d['s1'] = stats.norm.rvs(*s.dpe_mean_std(
            ndet=d['photon_detected'],
            p_dpe=gimme('double_pe_fraction'),
            mean_per_q=gimme('photon_gain_mean'),
            std_per_q=gimme('photon_gain_std')))

        acceptance = np.ones(len(d))
        for q in quanta_types:
            acceptance *= gimme(q + '_acceptance', d[q + '_detected'].values)
            sn = signal_name[q]
            acceptance *= gimme(sn + '_acceptance', d[sn].values)
        d = d.iloc[np.random.rand(len(d)) < acceptance].copy()

        # Now that we have s1 and s2 values, we can do the full annotate,
        # populating columns like e_vis, photon_produced_mle, etc.
        cls.annotate_data(d, **params)
        assert 'e_vis' in d.columns
        return d

    def _simulate_nq(self):
        raise NotImplementedError

    @classmethod
    def mu_interpolator(cls,
                        data,
                        interpolation_method='star',
                        n_trials=int(1e5),
                        **params):
        """Return interpolator for number of expected events
        Parameters must be specified as kwarg=(start, stop, n_anchors)
        """
        # TODO: is the mu also to be batched?
        if interpolation_method != 'star':
            raise NotImplementedError(
                f"mu interpolation method {interpolation_method} "
                f"not implemented")

        base_mu = tf.constant(cls.estimate_mu(data, n_trials=n_trials),
                              dtype=fd.float_type())
        pspaces = dict()    # parameter -> tf.linspace of anchors
        mus = dict()        # parameter -> tensor of mus
        for pname, pspace_spec in tqdm(params.items(),
                                       desc="Estimating mus"):
            pspaces[pname] = tf.linspace(*pspace_spec)
            mus[pname] = tf.convert_to_tensor(
                 [cls.estimate_mu(data, **{pname: x}, n_trials=n_trials)
                 for x in np.linspace(*pspace_spec)],
                dtype=fd.float_type())

        def mu_itp(**kwargs):
            mu = base_mu
            for pname, v in kwargs.items():
                mu *= tfp.math.interp_regular_1d_grid(
                    x=v,
                    x_ref_min=params[pname][0],
                    x_ref_max=params[pname][1],
                    y_ref=mus[pname]) / base_mu
            return mu

        return mu_itp

    @classmethod
    def estimate_mu(cls, data=None, n_trials=int(1e5), **params):
        """Return estimate of total expected number of events
        :param data: Data used for drawing auxiliary observables
        (e.g. position and time)
        :param n_trials: Number of events to simulate for efficiency estimate
        """
        # TODO what if e_spectrum is pos/time dependent?
        _, spectra = cls(data, _skip_tf_init=True).gimme(
            'energy_spectrum', numpy_out=True)
        mean_rate = spectra.sum(axis=1).mean(axis=0)

        d_simulated = cls.simulate(n_trials, data=data, **params)
        return mean_rate * len(d_simulated) / n_trials



from multihist import Hist1d
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from scipy import stats

import flamedisx as fd
export, __all__ = fd.exporter()

from .source import data_methods, special_data_methods

o = tf.newaxis


@export
class ERSource(fd.Source):

    work = 13.7e-3

    def energy_spectrum(self, drift_time):
        """Return (energies in keV, rate at these energies),
        both (n_events, n_energies) tensors.
        """
        # TODO: doesn't depend on drift_time...
        n_evts = len(drift_time)
        return (fd.repeat(tf.cast(tf.linspace(0., 10., 1000)[o, :],
                                 dtype=fd.float_type()),
                         n_evts, axis=0),
                fd.repeat(tf.ones(1000, dtype=fd.float_type())[o, :],
                         n_evts, axis=0))

    def energy_spectrum_hist(self):
        # TODO: fails if e is pos/time dependent
        es, rs = self.gimme('energy_spectrum', numpy_out=True)
        return Hist1d.from_histogram(rs[0, :-1], es[0, :])

    def simulate_es(self, n):
        return self.energy_spectrum_hist().get_random(n)

    @staticmethod
    def p_electron(nq):
        return 0.5 * tf.ones_like(nq, dtype=fd.float_type())

    @staticmethod
    def p_electron_fluctuation(nq):
        return 0.01 * tf.ones_like(nq, dtype=fd.float_type())

    @staticmethod
    def penning_quenching_eff(nph):
        return tf.ones_like(nph, dtype=fd.float_type())

    # Detection efficiencies

    @staticmethod
    def electron_detection_eff(drift_time, *, elife=600e3, extraction_eff=1.):
        return extraction_eff * tf.exp(-drift_time / elife)

    photon_detection_eff = 0.1

    # Acceptance of selection/detection on photons/electrons detected
    # The min_xxx attributes are also used in the bound computations
    min_s1_photons_detected = 3.
    min_s2_electrons_detected = 3.

    def electron_acceptance(self, electrons_detected):
        return tf.where(
            electrons_detected < self.min_s2_electrons_detected,
            tf.zeros_like(electrons_detected, dtype=fd.float_type()),
            tf.ones_like(electrons_detected, dtype=fd.float_type()))

    def photon_acceptance(self, photons_detected):
        return tf.where(
            photons_detected < self.min_s1_photons_detected,
            tf.zeros_like(photons_detected, dtype=fd.float_type()),
            tf.ones_like(photons_detected, dtype=fd.float_type()))

    # Acceptance of selections on S1/S2 directly

    @staticmethod
    def s1_acceptance(s1):
        return tf.where(s1 < 2,
                        tf.zeros_like(s1, dtype=fd.float_type()),
                        tf.ones_like(s1, dtype=fd.float_type()))

    @staticmethod
    def s2_acceptance(s2):
        return tf.where(s2 < 200,
                        tf.zeros_like(s2, dtype=fd.float_type()),
                        tf.ones_like(s2, dtype=fd.float_type()))

    electron_gain_mean = 20.
    electron_gain_std = 5.

    photon_gain_mean = 1.
    photon_gain_std = 0.5
    double_pe_fraction = 0.219

    def _simulate_nq(self):
        work = self.gimme('work', numpy_out=True)
        self.data['nq'] = np.floor(self.data['energy'].values / work).astype(np.int)


@export
class NRSource(ERSource):
    # TODO: needs batching as well!!
    do_pel_fluct = False
    data_methods = tuple(data_methods + ['lindhard_l'])
    special_data_methods = tuple(special_data_methods + ['lindhard_l'])

    @staticmethod
    def lindhard_l(e, lindhard_k=0.138):
        """Return Lindhard quenching factor at energy e in keV"""
        eps = 11.5 * e * 54**(-7/3)             # Xenon: Z = 54
        g = 3. * eps**0.15 + 0.7 * eps**0.6 + eps
        res = lindhard_k * g/(1. + lindhard_k * g)
        return res

    def energy_spectrum(self, drift_time):
        """Return (energies in keV, events at these energies),
        both (n_events, n_energies) tensors.
        """
        e = fd.repeat(tf.cast(tf.linspace(0.7, 150., 100)[o, :],
                              fd.float_type()),
                      len(drift_time), axis=0)
        return e, tf.ones_like(e, dtype=fd.float_type())

    def rate_nq(self, nq_1d, i_batch=None):
        # (n_events, |ne|) tensors
        es, rate_e = self.gimme('energy_spectrum', i_batch=i_batch)
        mean_q_produced = (
                es
                * self.gimme('lindhard_l', es, i_batch=i_batch)
                / self.gimme('work', i_batch=i_batch)[:, o])

        # (n_events, |nq|, |ne|) tensor giving p(nq | e)
        p_nq_e = tfp.distributions.Poisson(
            mean_q_produced[:, o, :]).prob(nq_1d[:, :, o])

        return tf.reduce_sum(p_nq_e * rate_e[:, o, :], axis=2)

    @staticmethod
    def penning_quenching_eff(nph, eta=8.2e-5 * 3.3, labda=0.8 * 1.15):
        return 1. / (1. + eta * nph ** labda)

    def _simulate_nq(self):
        work = self.gimme('work', numpy_out=True)
        lindhard_l = self.gimme('lindhard_l', self.data['energy'].values,
                                numpy_out=True)
        self.data['nq'] = stats.poisson.rvs(
            self.data['energy'].values * lindhard_l / work)
